import doctest
import io
from contextlib import redirect_stderr, redirect_stdout
from textwrap import dedent

from okgrade.result import TestResult

class SingleDocTest:
    """
    A single DocTest.

    Instances of this class are callable. When called, it takes 
    a global_environment dict, and returns a TestResult object.

    We only take a global_environment, *not* a local_environment.
    This makes tests not useful inside functions, methods or
    other scopes with local variables. This is a limitation of
    doctest, so we roll with it.
    """
    def __init__(self, name, doctest_string):
        self.name = name
        self.doctest_string = doctest_string
        self.examples = doctest.DocTestParser().parse(
            doctest_string,
            name
        )

    PLAIN_TEXT_SUMMARY_TEMPLATE = dedent(r"""
    Test {name} failed!

    Test code:
    {doctest_string}

    Test result:
    {runresults}
    """).strip()

    def __call__(self, global_environment):
        """
        Run test with given global_environment.
        """
        test = doctest.DocTest(
            [e for e in self.examples if isinstance(e, doctest.Example)],
            global_environment,
            self.name,
            None,
            None,
            self.doctest_string
        )

        doctestrunner = doctest.DocTestRunner(verbose=True)

        runresults = io.StringIO()
        with redirect_stdout(runresults), redirect_stderr(runresults):
            doctestrunner.run(test, clear_globs=False)
        with open('/dev/null', 'w') as f, redirect_stderr(f), redirect_stdout(f):
            result = doctestrunner.summarize(verbose=True)
        grade = 1.0 - (result.failed / result.attempted)
        if grade == 1.0:
            summary = 'Test {} passed!'.format(self.name)
        else:
            summary = self.PLAIN_TEXT_SUMMARY_TEMPLATE.format(
                name=self.name,
                doctest_string=dedent(self.doctest_string),
                runresults=runresults.getvalue()
            )
        return TestResult(grade, summary)


import html

class TestResult:
    """
    The result of running a test against an environment.

    Contains a grade (float) and a dictionary of 'summaries'
    that can be displayed to the user, in various mime types.
    """
    def __init__(self, grade, summary_mimebundle):
        """
        Result of running a Test of some kind.

        grade - float, the grade this test produced.
        summary_mimebundle - dict | string, mapping mimetypes to summaries.
                             If string is passed in, assume it is for mimetype 
                             text/plain.
        """
        self.grade = grade
        if isinstance(summary_mimebundle, str):
            self.summary_mimebundle = {'text/plain': summary_mimebundle}
        else:
            if 'text/plain' not in summary_mimebundle:
                raise ValueError('summary_mimebundle must contain text/plain')
            self.summary_mimebundle = summary_mimebundle

    def get_summary(self, mimetype='text/plain'):
        """
        Return summary for this TestResult.

        If mimetype is not passed in, 'text/plain' is assumed
        """
        return self.summary_mimebundle[mimetype]

    def _repr_html_(self):
        """
        Return HTML representation of this Test Result.

        Used by IPython to display pretty results
        """
        if 'text/html' in self.summary_mimebundle:
            return self.get_summary('text/html')
        else:
            return '<pre>' + html.escape(self.get_summary('text/plain')) + '</pre>'

    def __eq__(self, other):
        if not isinstance(other, TestResult):
            raise ValueError('Can not compare TestResult object with object of type {}'.format(type(other)))

        return other.grade == self.grade and other.summary_mimebundle == self.summary_mimebundle


from setuptools import find_packages, setup

setup(
    name='okgrade',
    version='0.4.1',
    license='3-clause BSD',
    author='Yuvi Panda',
    author_email='yuvipanda@gmail.com',
    description='Simple alternative to okpy for grading only',
    packages=find_packages()
)



from contextlib import redirect_stderr, redirect_stdout
import json
import inspect
from okgrade.grader import grade
from okgrade.result import TestResult

try:
    from IPython.core.inputsplitter import IPythonInputSplitter
except ImportError:
    raise ImportError('IPython needs to be installed for notebook grading')

def execute_notebook(nb, initial_env=None, ignore_errors=False):
    """
    Execute notebook & return the global environment that results from execution.

    If ignore_errors is True, exceptions are swallowed.

    nb is passed in as a dictionary that's a parsed ipynb file
    """
    if initial_env:
        global_env = initial_env.copy()
    else:
        global_env = {}
    for cell in nb['cells']:
        if cell['cell_type'] == 'code':
            # transform the input to executable Python
            # FIXME: use appropriate IPython functions here
            isp = IPythonInputSplitter(line_input_checker=False)
            # FIXME: We just strip out all magics for now
            source = isp.transform_cell('\n'.join(l for l in cell['source'] if not l.startswith('%')))
            try:
                with open('/dev/null', 'w') as f, redirect_stdout(f), redirect_stderr(f):
                    exec(source, global_env)
            except:
                if not ignore_errors:
                    raise
    return global_env

def _global_anywhere(varname):
    """
    Return global with given name in any frame in the call stack

    Throws NameError if no such global exists anywhere in the call stack
    """
    # This should not be a recursive function, since that modifies the stack!
    cur_frame = inspect.currentframe().f_back
    while cur_frame is not None:
        if varname in cur_frame.f_globals:
            return cur_frame.f_globals[varname]
        cur_frame = cur_frame.f_back
    raise NameError(f'{varname} not found in any globals in the stack')


def grade_notebook(notebook_path, test_files, ignore_errors=True):
    """
    Grade a notebook file & return grade
    """
    try:
        # Lots of notebooks call grade_notebook in them. These notebooks are then
        # executed by okgrade - which will in-turn execute grade_notebook again!
        # This puts us in an infinite loop.
        # We use this sentinel to detect and break out of that loop.
        _global_anywhere('__OKGRADE__')
        # FIXME: Do something else here?
        return None
    except NameError:
        pass

    with open(notebook_path) as f:
        nb = json.load(f)

    initial_env = {
        # Set this to prevent recursive executions!
        '__OKGRADE__': True
    }

    global_env = execute_notebook(nb, initial_env, ignore_errors=ignore_errors)

    # FIXME: This needs to be more general
    results = [grade(tf, global_env) for tf in test_files]
    
    test_grade = sum([r.grade for r in results]) / len(results)
    if test_grade == 1:
        return TestResult(1.0, 'Grade is: 100%')
    else:
        # FIXME: This is terrible!
        grade_pct = test_grade * 100
        summary = f"Grade is: {grade_pct}%\n"
        summary += "\n".join([r.get_summary() for r in results])
        return TestResult(test_grade, summary)



import inspect
from okgrade.doctest import SingleDocTest
from okgrade.result import TestResult
try:
    from IPython.core.formatters import BaseFormatter

    class NullFormatter(BaseFormatter):
        def __call__(self, obj):
            return
except NameError:
    pass

def parse_ok_test(path):
    """
    Parse a ok test file & return list of SingleDocTests.
    """
    # ok test files are python files, with a global 'test' defined
    test_globals = {}
    with open(path) as f:
        exec(f.read(), test_globals)
    
    test_spec = test_globals['test']

    # We only support a subset of these tests, so let's validate!

    # Make sure there is a name
    assert 'name' in test_spec

    # Do not support multiple suites in the same file
    assert len(test_spec['suites']) == 1

    # Do not support point values other than 1
    assert test_spec.get('points', 1) == 1

    test_suite = test_spec['suites'][0]

    # Only support doctest. I am unsure if other tests are implemented
    assert test_suite.get('type', 'doctest') == 'doctest'

    # Not setup and teardown supported
    assert not bool(test_suite.get('setup'))
    assert not bool(test_suite.get('teardown'))

    tests = []

    for i, test_case in enumerate(test_spec['suites'][0]['cases']):
        tests.append(SingleDocTest(
            test_spec['name'] + ' ' + str(i + 1),
            test_case['code']
        ))

    return tests


def grade(test_file_path, global_env=None):
    """
    Grade global_env against given test_file

    If global_env is none, the global environment of the calling
    function is used. The following two calls are equivalent:

    grade('somefile.ok')

    grade('somefile.ok', globals())

    Returns a TestResult object.
    """
    tests = parse_ok_test(test_file_path)
    if global_env is None:
        # Get the global env of our callers - one level below us in the stack
        # The grade method should only be called directly from user / notebook
        # code. If some other method is calling it, it should also use the
        # inspect trick to pass in its parents' global env.
        global_env = inspect.currentframe().f_back.f_globals
    try:
        from IPython import get_ipython
        ipy = get_ipython()
    except NameError:
        ipy = None
    for test in tests:
        try:
            if ipy is not None:
                old_formatters = ipy.display_formatter.formatters
                ipy.display_formatter.formatters = {'text/plain': NullFormatter()}
            resp = test(global_env)
        finally:
            if ipy is not None:
                ipy.display_formatter.formatters = old_formatters
        if resp.grade == 0:
            return resp
    # All tests passed!
    return TestResult(1, f"{test_file_path}: All tests passed!")